import json
import sys
from pathlib import Path
from shutil import rmtree
from subprocess import DEVNULL, PIPE, CalledProcessError, run  # nosec
from typing import Any, Dict

from cookiecutter.generate import generate_files
from examples import example
from git import Repo

from cruft.commands._utils import (
    RobustTemporaryDirectory,
    generate_cookiecutter_context,
    get_cookiecutter_repo,
    get_cruft_file,
    json_dumps,
)

try:
    import toml  # type: ignore
except ImportError:  # pragma: no cover
    toml = None  # type: ignore

CruftState = Dict[str, Any]


@example()
@example(skip_apply_ask=True)
def update(
    expanded_dir: str = ".",
    cookiecutter_input: bool = False,
    skip_apply_ask: bool = False,
    skip_update: bool = False,
    checkout: str = None,
) -> bool:
    """Update specified project's cruft to the latest and greatest release."""
    expanded_dir_path = Path(expanded_dir)
    pyproject_file = expanded_dir_path / "pyproject.toml"
    cruft_file = get_cruft_file(expanded_dir_path)

    cruft_state = json.loads(cruft_file.read_text())

    with RobustTemporaryDirectory() as compare_directory_str:
        # Initial setup
        compare_directory = Path(compare_directory_str)
        template_dir = compare_directory / "template"
        repo = get_cookiecutter_repo(cruft_state["template"], template_dir, checkout)
        directory = cruft_state.get("directory", "")
        if directory:
            template_dir = template_dir / directory
        last_commit = repo.head.object.hexsha

        # Bail early if the repo is already up to date
        if last_commit == cruft_state["commit"] or not repo.index.diff(cruft_state["commit"]):
            return False

        # Generate clean outputs via the cookiecutter
        # from the current cruft state commit of the cookiectter and the updated
        # cookiecutter.
        old_main_directory, new_main_directory, new_context = _generate_project_updates(
            compare_directory, cruft_state, template_dir, cookiecutter_input, repo
        )

        # Remove any files from the generated versions that we are supposed
        # to skip before generating the diff and applying updates
        _remove_skip_files(cruft_state, pyproject_file, old_main_directory, new_main_directory)

        # Given the two versions of the cookiecutter outputs based
        # on the current project's context we calculate the diff and
        # apply the updates to the current project.
        _apply_project_updates(
            old_main_directory, new_main_directory, expanded_dir_path, skip_update, skip_apply_ask
        )

        # Update the cruft state and dump the new state
        # to the cruft file
        cruft_state["commit"] = last_commit
        cruft_state["context"] = new_context
        cruft_state["directory"] = directory
        cruft_file.write_text(json_dumps(cruft_state))

        return True


#####################################
# Generating clean outputs for diff #
#####################################


def _generate_output(
    cruft_state: CruftState, template_dir: Path, cookiecutter_input: bool, new_output_dir: Path
):
    new_context = generate_cookiecutter_context(
        cruft_state["template"],
        template_dir,
        extra_context=cruft_state["context"]["cookiecutter"],
        no_input=not cookiecutter_input,
    )

    project_dir = generate_files(
        repo_dir=template_dir,
        context=new_context,
        overwrite_if_exists=True,
        output_dir=new_output_dir,
    )
    return new_context, Path(project_dir)


def _generate_project_updates(
    compare_directory: Path,
    cruft_state: CruftState,
    template_dir: Path,
    cookiecutter_input: bool,
    repo: Repo,
):
    new_output_dir = compare_directory / "new_output"

    new_context, new_main_directory = _generate_output(
        cruft_state, template_dir, cookiecutter_input, new_output_dir
    )

    repo.head.reset(commit=cruft_state["commit"], working_tree=True)

    old_output_dir = compare_directory / "old_output"
    _, old_main_directory = _generate_output(
        cruft_state, template_dir, cookiecutter_input, old_output_dir
    )
    return old_main_directory, new_main_directory, new_context


##############################
# Removing unnecessary files #
##############################


def _remove_skip_files(
    cruft_state: CruftState,
    pyproject_file: Path,
    old_main_directory: Path,
    new_main_directory: Path,
):
    skip_cruft = cruft_state.get("skip", [])
    if toml and pyproject_file.is_file():
        pyproject_cruft = toml.loads(pyproject_file.read_text()).get("tool", {}).get("cruft", {})
        skip_cruft.extend(pyproject_cruft.get("skip", []))

    for skip_file in skip_cruft:
        file_path_old = old_main_directory / skip_file
        file_path_new = new_main_directory / skip_file
        for file_path in (file_path_old, file_path_new):
            if file_path.is_dir():
                rmtree(file_path)
            elif file_path.is_file():
                file_path.unlink()


#################################################
# Calculating project diff and applying updates #
#################################################


def _get_diff(old_main_directory: Path, new_main_directory: Path):
    diff = run(
        [
            "git",
            "diff",
            "--no-index",
            "--no-ext-diff",
            "--no-color",
            str(old_main_directory),
            str(new_main_directory),
        ],
        stdout=PIPE,
        stderr=PIPE,
    ).stdout.decode("utf8")
    diff = (
        diff.replace("\\\\", "\\")
        .replace(str(old_main_directory), "")
        .replace(str(new_main_directory), "")
    )
    return diff


def _view_diff(old_main_directory: Path, new_main_directory: Path):
    run(["git", "diff", "--no-index", str(old_main_directory), str(new_main_directory)])


def _is_git_repo(directory: Path):
    # Taken from https://stackoverflow.com/a/16925062
    # This works even if we are in a sub folder in a git
    # repo
    output = run(
        ["git", "rev-parse", "--is-inside-work-tree"], stdout=PIPE, stderr=DEVNULL, cwd=directory
    )
    if b"true" in output.stdout:
        return True
    return False


def _apply_patch(diff: str, expanded_dir_path: Path):
    # Git 3 way merge is the our best bet
    # at applying patches. But it only works
    # with git repos. If the repo is not a git dir
    # we fall back to git apply --reject which applies
    # diffs cleanly where applicable otherwise creates
    # *.rej files where there are conflicts
    if _is_git_repo(expanded_dir_path):
        patch_command = ["git", "apply", "-3"]
    else:
        patch_command = ["git", "apply", "--reject"]
    try:
        run(
            patch_command, input=diff.encode("utf8"), stderr=PIPE, check=True, cwd=expanded_dir_path
        )
    except CalledProcessError as error:
        print(error.stderr.decode(), file=sys.stderr)


def _apply_project_updates(
    old_main_directory: Path,
    new_main_directory: Path,
    project_dir: Path,
    skip_update: bool,
    skip_apply_ask: bool,
):
    diff = _get_diff(old_main_directory, new_main_directory)

    if not skip_apply_ask and not skip_update:  # pragma: no cover
        input_str: str = ""
        while input_str not in ("y", "n", "s"):
            print(
                'Respond with "s" to intentionally skip the update while marking '
                "your project as up-to-date or "
                'respond with "v" to view the changes that will be applied.'
            )
            input_str = input("Apply diff and update [y/n/s/v]? ").lower()  # nosec
            if input_str == "v":
                if diff.strip():
                    _view_diff(old_main_directory, new_main_directory)
                else:
                    print("There are no changes.")
        if input_str == "n":
            sys.exit("User cancelled Cookiecutter template update.")
        elif input_str == "s":
            skip_update = True

    if not skip_update:
        _apply_patch(diff, project_dir)



import json
from pathlib import Path

from examples import example

from cruft.commands._utils import RobustTemporaryDirectory, get_cookiecutter_repo, get_cruft_file


@example()
def check(expanded_dir: str = ".", checkout: str = None) -> bool:
    """Checks to see if there have been any updates to the Cookiecutter template used
    to generate this project.
    """
    expanded_dir_path = Path(expanded_dir)
    cruft_file = get_cruft_file(expanded_dir_path)
    cruft_state = json.loads(cruft_file.read_text())
    with RobustTemporaryDirectory() as cookiecutter_template_dir:
        repo = get_cookiecutter_repo(cruft_state["template"], cookiecutter_template_dir, checkout)
        last_commit = repo.head.object.hexsha
        if last_commit == cruft_state["commit"] or not repo.index.diff(cruft_state["commit"]):
            return True

    return False



"""This module defines CLI interaction when using `cruft`.

This is powered by [hug](https://github.com/hugapi/hug) which means unless necessary
it should maintain 1:1 compatibility with the programmatic API definition in the
[API module](/reference/cruft/api)

- `cruft create`: Expands the specified Cookiecutter template on disk.
- `cruft check`: Checks to see if the expanded template is up-to-date with latest version.
- `cruft update`: Attempts to updates an expanded Cookiecutter template to the latest version.
"""
import sys

import hug

from cruft import api, logo

hug_api = hug.API(__name__, doc=logo.ascii_art)


def _check_command_output(up_to_date: bool) -> None:
    if not up_to_date:
        print("")
        sys.exit(
            "FAILURE: Project's cruft is out of date! Run `cruft update` to clean this mess up."
        )
    else:
        print("")
        print("SUCCESS: Good work! Project's cruft is up to date and as clean as possible :).")


def _update_output(updated: bool) -> None:
    if not updated:
        print("")
        print("Nothing to do, project's cruft is already up to date!")
    else:
        print("")
        print("Good work! Project's cruft has been updated and is as clean as possible!")


def _link_output(linked: bool) -> None:
    if linked:
        print("")
        print("Project successfully linked to template!")
    else:
        print("")
        print("Project linking failed :(")


cli = hug.cli(api=hug_api)
cli(api.create)
cli.output(_update_output)(api.update)
cli.output(_check_command_output)(api.check)
cli.output(_link_output)(api.link)



from pathlib import Path

from examples import example

from cruft.commands._utils import (
    RobustTemporaryDirectory,
    generate_cookiecutter_context,
    get_cookiecutter_repo,
    get_cruft_file,
    json_dumps,
)


@example("https://github.com/timothycrosley/cookiecutter-python/", no_input=True, use_latest=True)
def link(
    template_git_url: str,
    project_dir: str = ".",
    use_latest: bool = False,
    no_input: bool = False,
    config_file: str = None,
    default_config: bool = False,
    extra_context: dict = None,
    directory: str = "",
) -> bool:
    """Links an existing project created from a template, to the template it was created from."""
    project_dir_path = Path(project_dir)
    cruft_file = get_cruft_file(project_dir_path, exists=False)

    with RobustTemporaryDirectory() as cookiecutter_template_dir_str:
        cookiecutter_template_dir = Path(cookiecutter_template_dir_str)
        repo = get_cookiecutter_repo(template_git_url, cookiecutter_template_dir)
        last_commit = repo.head.object.hexsha

        if directory:
            cookiecutter_template_dir = cookiecutter_template_dir / directory

        context = generate_cookiecutter_context(
            template_git_url,
            cookiecutter_template_dir,
            config_file,
            default_config,
            extra_context,
            no_input,
        )

        if use_latest or no_input:
            use_commit = last_commit
        else:  # pragma: no cover
            print("")
            print(f"The latest commit to the template is {last_commit}")
            print("Press enter to link against this commit or provide an alternative commit.")
            print("")
            use_commit = input(f"Link to template at commit [{last_commit}]: ")  # nosec
            use_commit = use_commit if use_commit.strip() else last_commit

        cruft_file.write_text(
            json_dumps(
                {
                    "template": template_git_url,
                    "commit": use_commit,
                    "context": context,
                    "directory": directory,
                }
            )
        )

    return True



from pathlib import Path

from cookiecutter.generate import generate_files
from examples import example

from cruft.commands._utils import (
    RobustTemporaryDirectory,
    generate_cookiecutter_context,
    get_cookiecutter_repo,
    json_dumps,
)


@example("https://github.com/timothycrosley/cookiecutter-python/", no_input=True)
def create(
    template_git_url: str,
    output_dir: str = ".",
    config_file: str = None,
    default_config: bool = False,
    extra_context: dict = None,
    no_input: bool = False,
    directory: str = "",
    checkout: str = None,
    overwrite_if_exists: bool = False,
) -> str:
    """Expand a Git based Cookiecutter template into a new project on disk."""
    with RobustTemporaryDirectory() as cookiecutter_template_dir_str:
        cookiecutter_template_dir = Path(cookiecutter_template_dir_str)
        repo = get_cookiecutter_repo(template_git_url, cookiecutter_template_dir, checkout)
        last_commit = repo.head.object.hexsha

        if directory:
            cookiecutter_template_dir = cookiecutter_template_dir / directory

        context = generate_cookiecutter_context(
            template_git_url,
            cookiecutter_template_dir,
            config_file,
            default_config,
            extra_context,
            no_input,
        )

        project_dir = generate_files(
            repo_dir=cookiecutter_template_dir,
            context=context,
            overwrite_if_exists=overwrite_if_exists,
            output_dir=output_dir,
        )

        # After generating the project - save the cruft state
        # into the cruft file.
        (Path(project_dir) / ".cruft.json").write_text(
            json_dumps(
                {
                    "template": template_git_url,
                    "commit": last_commit,
                    "context": context,
                    "directory": directory,
                }
            )
        )

        return project_dir



from cruft.commands.check import check
from cruft.commands.create import create
from cruft.commands.link import link
from cruft.commands.update import update

__all__ = ["check", "create", "link", "update"]



import json
import os
import stat
import time
from functools import partial
from pathlib import Path
from shutil import rmtree
from tempfile import TemporaryDirectory
from typing import Any, Dict, Optional

from cookiecutter.config import get_user_config
from cookiecutter.generate import generate_context
from cookiecutter.prompt import prompt_for_config
from git import Repo

from cruft.exceptions import (
    CruftAlreadyPresent,
    InvalidCookiecutterRepository,
    NoCruftFound,
    UnableToFindCookiecutterTemplate,
)

#################################
# Cookiecutter helper functions #
#################################


def get_cookiecutter_repo(
    template_git_url: str, cookiecutter_template_dir: Path, checkout: Optional[str] = None
):
    try:
        repo = Repo.clone_from(template_git_url, cookiecutter_template_dir)
        if checkout is not None:
            repo.git.checkout(checkout)
    except Exception as e:  # pragma: no cover
        raise InvalidCookiecutterRepository(e)
    else:
        return repo


def _validate_cookiecutter(cookiecutter_template_dir: Path):
    main_cookiecutter_directory: Optional[Path] = None

    for dir_item in cookiecutter_template_dir.glob("*cookiecutter.*"):
        if dir_item.is_dir() and "{{" in dir_item.name and "}}" in dir_item.name:
            main_cookiecutter_directory = dir_item
            break

    if not main_cookiecutter_directory:  # pragma: no cover
        raise UnableToFindCookiecutterTemplate(cookiecutter_template_dir)


def generate_cookiecutter_context(
    template_git_url: str,
    cookiecutter_template_dir: Path,
    config_file: str = None,
    default_config: bool = False,
    extra_context: Dict[str, Any] = None,
    no_input: bool = False,
):
    _validate_cookiecutter(cookiecutter_template_dir)

    context_file = cookiecutter_template_dir / "cookiecutter.json"

    config_dict = get_user_config(config_file=config_file, default_config=default_config)

    context = generate_context(
        context_file=context_file,
        default_context=config_dict["default_context"],
        extra_context=extra_context,
    )

    # prompt the user to manually configure at the command line.
    # except when 'no-input' flag is set
    context["cookiecutter"] = prompt_for_config(context, no_input)
    context["cookiecutter"]["_template"] = template_git_url

    return context


#######################
# Cruft related utils #
#######################


def get_cruft_file(project_dir_path: Path, exists: bool = True):
    cruft_file = project_dir_path / ".cruft.json"
    if not exists and cruft_file.is_file():
        raise CruftAlreadyPresent(cruft_file)
    if exists and not cruft_file.is_file():
        raise NoCruftFound(project_dir_path.resolve())
    return cruft_file


class RobustTemporaryDirectory(TemporaryDirectory):
    """Retries deletion on __exit__
    This is caused by Windows behavior that you cannot delete a directory
    if it contains any read-only files.
    cf. https://bugs.python.org/issue19643
    """

    DELETE_MAX_RETRY_COUNT = 10
    DELETE_RETRY_TIME = 0.1

    def cleanup(self):
        if self._finalizer.detach():

            def readonly_handler(rm_func, path, exc_info):
                if issubclass(exc_info[0], PermissionError):
                    os.chmod(path, stat.S_IWRITE)
                    return rm_func(path)

            err_count = 0
            while True:
                try:
                    rmtree(self.name, onerror=readonly_handler)
                    break
                except (OSError, WindowsError):
                    err_count += 1
                    if err_count > self.DELETE_MAX_RETRY_COUNT:
                        # This serves as a workaround to be able to use this tool under Windows.
                        # Deleting temporary folders fails because Python cannot delete them.
                        if os.name != "nt":
                            raise
                        else:
                            break
                    time.sleep(self.DELETE_RETRY_TIME)


json_dumps = partial(json.dumps, ensure_ascii=False, indent=4, separators=(",", ": "))



"""**cruft**

Allows you to maintain all the necessary cruft for packaging and building projects separate from
the code you intentionally write. Built on-top of, and fully compatible with, CookieCutter.
"""
from cruft._version import __version__
